// Copyright (c) 2007 Keith D Gregory
package com.kdgregory.example.ref.phantom;

import java.sql.Array;
import java.sql.Blob;
import java.sql.CallableStatement;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.NClob;
import java.sql.PreparedStatement;
import java.sql.SQLClientInfoException;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.SQLXML;
import java.sql.Savepoint;
import java.sql.Statement;
import java.sql.Struct;
import java.util.Map;
import java.util.Properties;

/**
 *  A wrapper for a JDBC database connection that is associated with a pool.
 *  Closing this connection returns the wrapped connection to the pool, to
 *  be used again.
 */
public class PooledConnection
implements Connection
{
    private ConnectionPool _pool;
    private Connection _cxt;


    public PooledConnection(ConnectionPool pool, Connection cxt)
    {
        _pool = pool;
        _cxt = cxt;
    }


    private Connection getConnection()
    {
        try
        {
            if ((_cxt == null) || _cxt.isClosed())
                throw new RuntimeException("Connection is closed");
        }
        catch (SQLException ex)
        {
            throw new RuntimeException("exception when testing if delegate is closed", ex);
        }
        return _cxt;
    }


//----------------------------------------------------------------------------
//  Implementation of Connection
//----------------------------------------------------------------------------

    @Override
    public void close() throws SQLException
    {
        if (_cxt != null)
        {
            _pool.releaseConnection(_cxt);
            _cxt = null;
        }
    }


    @Override
    public boolean isClosed() throws SQLException
    {
        return (_cxt == null) || (_cxt.isClosed());
    }


    @Override
    public void clearWarnings() throws SQLException
    {
        getConnection().clearWarnings();
    }


    @Override
    public void commit() throws SQLException
    {
        getConnection().commit();
    }


    @Override
    public Statement createStatement() throws SQLException
    {
        return getConnection().createStatement();
    }


    @Override
    public Statement createStatement(int resultSetType, int resultSetConcurrency, int resultSetHoldability) throws SQLException
    {
        return getConnection()
               .createStatement(resultSetType, resultSetConcurrency, resultSetHoldability);
    }


    @Override
    public Statement createStatement(int resultSetType, int resultSetConcurrency) throws SQLException
    {
        return getConnection()
               .createStatement(resultSetType, resultSetConcurrency);
    }


    @Override
    public boolean getAutoCommit() throws SQLException
    {
        return getConnection().getAutoCommit();
    }


    @Override
    public String getCatalog() throws SQLException
    {
        return getConnection().getCatalog();
    }


    @Override
    public int getHoldability() throws SQLException
    {
        return getConnection().getHoldability();
    }


    @Override
    public DatabaseMetaData getMetaData() throws SQLException
    {
        return getConnection().getMetaData();
    }


    @Override
    public int getTransactionIsolation() throws SQLException
    {
        return getConnection().getTransactionIsolation();
    }


    @Override
    public Map<String,Class<?>> getTypeMap() throws SQLException
    {
        return getConnection().getTypeMap();
    }


    @Override
    public SQLWarning getWarnings() throws SQLException
    {
        return getConnection().getWarnings();
    }


    @Override
    public boolean isReadOnly() throws SQLException
    {
        return getConnection().isReadOnly();
    }


    @Override
    public String nativeSQL(String sql) throws SQLException
    {
        return getConnection().nativeSQL(sql);
    }


    @Override
    public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency, int resultSetHoldability) throws SQLException
    {
        return getConnection()
               .prepareCall(sql, resultSetType, resultSetConcurrency, resultSetHoldability);
    }


    @Override
    public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency) throws SQLException
    {
        return getConnection()
               .prepareCall(sql, resultSetType, resultSetConcurrency);
    }


    @Override
    public CallableStatement prepareCall(String sql) throws SQLException
    {
        return getConnection().prepareCall(sql);
    }


    @Override
    public PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency, int resultSetHoldability) throws SQLException
    {
        return getConnection()
               .prepareStatement(sql, resultSetType, resultSetConcurrency, resultSetHoldability);
    }


    @Override
    public PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency) throws SQLException
    {
        return getConnection()
               .prepareStatement(sql, resultSetType, resultSetConcurrency);
    }


    @Override
    public PreparedStatement prepareStatement(String sql, int autoGeneratedKeys) throws SQLException
    {
        return getConnection()
               .prepareStatement(sql, autoGeneratedKeys);
    }


    @Override
    public PreparedStatement prepareStatement(String sql, int[] columnIndexes) throws SQLException
    {
        return getConnection().prepareStatement(sql, columnIndexes);
    }


    @Override
    public PreparedStatement prepareStatement(String sql, String[] columnNames) throws SQLException
    {
        return getConnection().prepareStatement(sql, columnNames);
    }


    @Override
    public PreparedStatement prepareStatement(String sql) throws SQLException
    {
        return getConnection().prepareStatement(sql);
    }


    @Override
    public void releaseSavepoint(Savepoint savepoint) throws SQLException
    {
        getConnection().releaseSavepoint(savepoint);
    }


    @Override
    public void rollback() throws SQLException
    {
        getConnection().rollback();
    }


    @Override
    public void rollback(Savepoint savepoint) throws SQLException
    {
        getConnection().rollback(savepoint);
    }


    @Override
    public void setAutoCommit(boolean autoCommit) throws SQLException
    {
        getConnection().setAutoCommit(autoCommit);
    }


    @Override
    public void setCatalog(String catalog) throws SQLException
    {
        getConnection().setCatalog(catalog);
    }


    @Override
    public void setHoldability(int holdability) throws SQLException
    {
        getConnection().setHoldability(holdability);
    }


    @Override
    public void setReadOnly(boolean readOnly) throws SQLException
    {
        getConnection().setReadOnly(readOnly);
    }


    @Override
    public Savepoint setSavepoint() throws SQLException
    {
        return getConnection().setSavepoint();
    }


    @Override
    public Savepoint setSavepoint(String name) throws SQLException
    {
        return getConnection().setSavepoint(name);
    }


    @Override
    public void setTransactionIsolation(int level) throws SQLException
    {
        getConnection().setTransactionIsolation(level);
    }


    @Override
    public void setTypeMap(Map<String,Class<?>> typemap) throws SQLException
    {
        getConnection().setTypeMap(typemap);
    }


    @Override
    public <T> T unwrap(Class<T> iface) throws SQLException
    {
        return getConnection().unwrap(iface);
    }


    @Override
    public boolean isWrapperFor(Class<?> iface) throws SQLException
    {
        return getConnection().isWrapperFor(iface);
    }


    @Override
    public Clob createClob() throws SQLException
    {
        return getConnection().createClob();
    }


    @Override
    public Blob createBlob() throws SQLException
    {
        return getConnection().createBlob();
    }


    @Override
    public NClob createNClob() throws SQLException
    {
        return getConnection().createNClob();
    }


    @Override
    public SQLXML createSQLXML() throws SQLException
    {
        return getConnection().createSQLXML();
    }


    @Override
    public boolean isValid(int timeout) throws SQLException
    {
        return getConnection().isValid(timeout);
    }


    @Override
    public void setClientInfo(String name, String value) throws SQLClientInfoException
    {
        getConnection().setClientInfo(name, value);
    }


    @Override
    public void setClientInfo(Properties properties) throws SQLClientInfoException
    {
        getConnection().setClientInfo(properties);
    }


    @Override
    public String getClientInfo(String name) throws SQLException
    {
        return getConnection().getClientInfo(name);
    }


    @Override
    public Properties getClientInfo() throws SQLException
    {
        return getConnection().getClientInfo();
    }


    @Override
    public Array createArrayOf(String typeName, Object[] elements) throws SQLException
    {
        return getConnection().createArrayOf(typeName, elements);
    }


    @Override
    public Struct createStruct(String typeName, Object[] attributes) throws SQLException
    {
        return getConnection().createStruct(typeName, attributes);
    }
}
